/*----------------------------------------------------------------------
  File    : TrainPanel.java
  Contents: panel for a train display (aka dot display)
  Author  : Christian Borgelt
  History : 2013.11.29 file created from file Chart3D.java
            2015.08.17 generally renamed point -> time
----------------------------------------------------------------------*/
package coconad;

import java.util.Arrays;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.image.BufferedImage;
import javax.swing.JPanel;

/*--------------------------------------------------------------------*/
/** Class for a train panel (aka dot display).
 *  @author Christian Borgelt
 *  @since  2013.11.29 */
/*--------------------------------------------------------------------*/
public class TrainPanel extends JPanel {

  private static final long serialVersionUID = 0x00010000L;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the height of a row for a train in pixels */
  private int      rowht;
  /** the scaling factor for the time dimension (horizontal axis) */
  private double   scale;
  /** the beginning of the time range to display */
  private double   beg;
  /** the end of the time range to display */
  private double   end;
  /** the color of the dots/lines for the events/spikes */
  private Color    color;
  /** the train set to display */
  private TrainSet trains;

  /*------------------------------------------------------------------*/
  /** Create a train panel.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainPanel ()
  {                             /* --- create a train panel */
    this.rowht = 2;             /*   3 pixels per train row */
    this.scale = 200.0;         /* 200 pixels per time unit */
    this.beg   = Double.NEGATIVE_INFINITY;
    this.end   = Double.POSITIVE_INFINITY;
    this.color = new Color(0.0F,0.0F,0.8F);
    this.setPreferredSize(new Dimension(600, 200));
  }  /* TrainPanel() */

  /*------------------------------------------------------------------*/
  /** Set the height of a row for a train.
   *  @param  ht the height of the row in pixels
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setRowHeight (int ht)
  { this.rowht = (ht > 0) ? ht : 1; }

  /*------------------------------------------------------------------*/
  /** Set the scaling factor for the time dimension (horizontal).
   *  @param  scale the number of pixels per time unit
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setScale (double scale)
  { this.scale = (scale > 0) ? scale : 1.0; }

  /*------------------------------------------------------------------*/
  /** Set the event/spike time range to display.
   *  @param  beg   the beginning of the time range to display
   *  @param  end   the end       of the time range to display
   *  @since  2013.11.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setRange (double beg, double end)
  { this.beg = beg; this.end = end; }

  /*------------------------------------------------------------------*/
  /** Set the color for the events/spikes.
   *  @param  color the color for the events/spikes
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setColor (Color color)
  { this.color = (color != null) ? color : Color.black; }

  /*------------------------------------------------------------------*/
  /** Get the current train set.
   *  @return the set of trains
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainSet getTrains ()
  { return this.trains; }

  /*------------------------------------------------------------------*/
  /** Get the current train set to display.
   *  @param  trains the current set of trains
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (TrainSet trains)
  { this.trains = trains; }

  /*------------------------------------------------------------------*/
  /** (Re)paint the train panel.
   *  @param  g the graphics to use for the painting
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void paint (Graphics g)
  {                             /* --- (re)paint the whole panel */
    int       i, k, n;          /* loop variables, buffer */
    Dimension d;                /* size of the panel */
    Train     train;            /* to traverse the trains */
    int       x, y, w, h;       /* coordinates of dot/line */
    double    min, max;         /* range of event/spike times */
    double    p, s;             /* time and scaling factor */

    super.paint(g);             /* ensure proper rendering queue */
    if (this.trains == null)    /* if there is no train set */
      min = max = s = n = 0;    /* clear all variables */
    else {                      /* if there is a train set */
      n   = this.trains.getSize();
      min = (this.beg > Double.NEGATIVE_INFINITY)
          ?  this.beg : Math.floor(this.trains.getMinTime());
      max = (this.end < Double.POSITIVE_INFINITY)
          ?  this.end : Math.ceil(this.trains.getMaxTime());
      if (min > Double.POSITIVE_INFINITY) {
        s = n = 0; this.setPreferredSize(new Dimension(600,200)); }
      else {                    /* if there are events to draw */
        s = (max-min < 65535/this.scale)
          ? this.scale : 65535.0/(max-min);
        w = (int)((max-min) *s +0.5);
        h = n *this.rowht;      /* compute the extensions */
        this.setPreferredSize(new Dimension(w,h));
        this.revalidate();      /* adapt the enclosing scroll pane */
      }                         /* set the computed dimenasions */
    }                           /* as the preferred panel size */
    d = this.getSize();         /* get the (preferred) panel size */
    w = d.width; h = d.height;  /* (whichever is larger) */
    d = this.getPreferredSize();
    if (d.width  > w) w = d.width;
    if (d.height > h) h = d.height;
    g.setColor(Color.white);    /* set the background color */
    g.fillRect(0, 0, w, h);;    /* and draw the background */
    if (n <= 0) return;         /* if there are no events, abort */
    g.setColor(this.color);     /* set the color of the dots/lines */
    for (i = n; --i >= 0;) {    /* traverse the trains */
      train = this.trains.getTrain(i);
      y = (n-i) *this.rowht;    /* get the y-coordinate */
      if (y >= d.height +this.rowht) continue;
      for (k = train.getLength(); --k >= 0; ) {
        p = train.getTime(k);   /* traverse the events of the train */
        if ((p < min) || (p > max)) continue;
        x = (int)((p -min) *s +0.5);
        g.drawLine(x, y-1, x, y-this.rowht);
      }                         /* draw a vertical line */
    }                           /* for each event/spike */
  }  /* paint() */

  /*------------------------------------------------------------------*/
  /** Create an image of the panel contents.
   *  @return an image of the panel contents
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public BufferedImage makeImage ()
  {                             /* --- create an image of contents */
    BufferedImage img;          /* created image */
    Dimension     d;            /* size of panel */

    d   = this.getPreferredSize();
    img = new BufferedImage(d.width, d.height,
                            BufferedImage.TYPE_INT_ARGB);
    this.paint(img.getGraphics());
    return img;                 /* draw window contents to image */
  }  /* BufferedImage() */      /* and return the image */

}  /* TrainPanel */
