/*----------------------------------------------------------------------
  File    : TrainView.java
  Contents: frame class for a train viewer
  Author  : Christian Borgelt
  History : 2013.11.29 file created from file BarChart.java
            2014.03.17 automatic choice of scaling factor added
            2014.10.23 changed from LGPL license to MIT license
            2015.08.17 generally renamed point -> time
----------------------------------------------------------------------*/
package coconad;

import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.BorderFactory;
import javax.imageio.ImageIO;

import util.TableReader;
import dialog.DialogPanel;
import dialog.FormatPanel;
import dialog.FormatDialog;
import dialog.ColorDialog;
import dialog.AboutDialog;

/*--------------------------------------------------------------------*/
/** Class for a train viewer frame.
 *  @author Christian Borgelt
 *  @since  2013.11.29 */
/*--------------------------------------------------------------------*/
public class TrainView extends JFrame implements Runnable {

  private static final long serialVersionUID = 0x00010002L;
  /** the version string */
  public  static final String VERSION = "1.2 (2014.10.23)";

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  /** mode flag: the scatter plot is a stand-alone program */
  public final static int PROGRAM    = 1;
  /** mode flag: add menu items for loading trains */
  public final static int LOAD_ITEMS = 2;

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the owner of this bar chart */
  private Component    owner   = null;
  /** the mode flags */
  private int          mode    = 0;
  /** the train view panel */
  private TrainPanel   panel   = null;
  /** the status bar for messages */
  private JTextField   status  = null;
  /** the file chooser */
  private JFileChooser chooser = null;
  /** the current trains file */
  private File         curr    = null;

  /* --- data format dialog box --- */
  /** the data format dialog box */
  private FormatDialog format  = null;

  /* --- layout dialog box --- */
  /** the layout dialog box */
  private JDialog      layout  = null;
  /** the height of each train row in pixels */
  private JSpinner     rowht   = null;
  /** the scaling factor for the horizontal dimension */
  private JTextField   scale   = null;
  /** the beginning of the time range to display */
  private JTextField   beg     = null;
  /** the end       of the time range to display */
  private JTextField   end     = null;

  /* --- color dialog box --- */
  /** the color selection dialog box */
  private ColorDialog  colsel  = null;

  /* --- about dialog box --- */
  /** the "About..." dialog box */
  private AboutDialog  about   = null;

  /*------------------------------------------------------------------*/
  /** Create a train view frame.
   *  @param  mode the mode flags
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainView (int mode)
  { this(null, mode); }

  /*------------------------------------------------------------------*/
  /** Create a train view frame.
   *  @param  owner the component that is to own this bar chart
   *  @param  mode  the mode flags
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainView (Component owner, int mode)
  {                             /* --- create a scatter plot frame */
    this.owner = owner;         /* note the owner */
    this.mode  = mode;          /* and the mode flags */
    if (EventQueue.isDispatchThread()) { this.run(); return; }
    try { EventQueue.invokeAndWait(this); }
    catch (Exception e) {}      /* create the user interface */
  }  /* TrainView() */

  /*------------------------------------------------------------------*/
  /** Create the user interface.
   *  <p>Following the recommendations in the Java tutorial, the user
   *  interface is created in the "run" method, which is invoked from
   *  the event queue, in order to avoid problems with threads.</p>
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void run ()
  {                             /* --- create a bar chart viewer */
    Container content;          /* content pane of this frame */
    JMenuBar  mbar;             /* menu bar */
    JMenu     menu;             /* to create menu titles */
    JMenuItem item;             /* to create menu items */

    if ((this.mode & PROGRAM) != 0)
      this.mode |= LOAD_ITEMS;  /* check and adapt the mode */

    /* --- create and set the menu bar --- */
    this.setJMenuBar(mbar = new JMenuBar());

    menu = mbar.add(new JMenu("File"));
    menu.setMnemonic('f');
    if ((this.mode & LOAD_ITEMS) != 0) {
      item = menu.add(new JMenuItem("Load Trains...", 'l'));
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          TrainView.this.loadTrains(null); } } );
      item = menu.add(new JMenuItem("Reload Trains", 'r'));
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          TrainView.this.loadTrains(TrainView.this.curr); } } );
      menu.addSeparator();
      item = menu.add(new JMenuItem("Data Format...", 'f'));
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          TrainView.this.getFormatDlg().setVisible(true);
          TrainView.this.format.toFront();
        } } );
      menu.addSeparator();
    }
    item = menu.add(new JMenuItem("Save PNG Image...", 'i'));
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.saveImage(null); } } );
    menu.addSeparator();
    item = menu.add(new JMenuItem("Quit", 'q'));
    if ((this.mode & PROGRAM) != 0) { /* if stand-alone program */
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          System.exit(0); } } ); }     /* terminate the program */
    else {                      /* if only visualization module */
      item.addActionListener(new ActionListener() {
        public void actionPerformed (ActionEvent e) {
          TrainView.this.setVisible(false); } } );
    }                           /* close the window */

    menu = mbar.add(new JMenu("View"));
    menu.setMnemonic('v');
    item = menu.add(new JMenuItem("Set Layout...", 'l'));
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.getLayoutDlg().setVisible(true);
        TrainView.this.layout.toFront();
      } } );
    item = menu.add(new JMenuItem("Set Color...", 'c'));
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.getColorDlg().setVisible(true);
        TrainView.this.colsel.toFront();
      } } );
    menu.addSeparator();
    item = menu.add(new JMenuItem("Redraw", 'r'));
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.panel.repaint(); } } );

    menu = mbar.add(new JMenu("Help"));
    menu.setMnemonic('h');
    item = menu.add(new JMenuItem("About...", 'a'));
    item.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        if (TrainView.this.about == null)
          TrainView.this.about = new AboutDialog(TrainView.this,
             "About TrainView...", "TrainView\n"
            +"A Train Set Viewer Program\n"
            +"Version " +TrainView.VERSION +"\n\n"
            +"written by Christian Borgelt\n"
            +"European Center for Soft Computing\n"
            +"c/ Gonzalo Gutierrez Quiros s/n\n"
            +"33600 Mieres, Asturias, Spain\n"
            +"christian@borgelt.net");
        TrainView.this.about.setVisible(true);
        TrainView.this.about.toFront();
      } } );

    /* --- create and set the main panel --- */
    this.panel = new TrainPanel();
    this.panel.setLayout(new BorderLayout());
    this.panel.setPreferredSize(new Dimension(600,200));
    content = this.getContentPane();
    content.setLayout(new BorderLayout());
    content.add(new JScrollPane(this.panel), BorderLayout.CENTER);

    /* --- create and set a status bar --- */
    this.status = new JTextField("");
    this.status.setEditable(false);
    content.add(this.status,  BorderLayout.SOUTH);

    /* --- configure and show the frame window --- */
    this.setTitle("TrainView");
    this.setDefaultCloseOperation(((this.mode & PROGRAM) != 0)
      ? JFrame.EXIT_ON_CLOSE : JFrame.HIDE_ON_CLOSE);
    if (this.owner == null) this.setLocation(48, 48);
    else                    this.setLocationRelativeTo(this.owner);
    this.pack();                /* configure the frame */
  }  /* run() */

  /*------------------------------------------------------------------*/
  /** Get the file chooser (create if necessary).
   *  @return the file chooser
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JFileChooser getFileChooser ()
  {                             /* --- get the file chooser */
    if (this.chooser != null)   /* if the chooser already exists, */
      return this.chooser;      /* simply return it */
    this.chooser = new JFileChooser();
    this.chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    this.chooser.setCurrentDirectory(new File("."));
    this.chooser.setFileHidingEnabled(true);
    this.chooser.setAcceptAllFileFilterUsed(true);
    this.chooser.setMultiSelectionEnabled(false);
    this.chooser.setFileView(null);
    return this.chooser;        /* create and configure the chooser */
  }  /* createChooser() */

  /*------------------------------------------------------------------*/
  /** Get the data format dialog (create if necessary).
   *  @return the data format dialog
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private FormatDialog getFormatDlg()
  {                             /* --- get the data format dialog */
    if (this.format == null)    /* if it does not exist, create it */
      this.format = new FormatDialog(this, "Set Data Format ...",
                                     FormatPanel.TRAIN);
    return this.format;         /* return the data format dialog */
  }  /* getFormatDlg() */

  /*------------------------------------------------------------------*/
  /** Set the layout.
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void setLayout ()
  {                             /* --- set the layout */
    double scale;               /* scaling factor */
    double beg, end;            /* beginning and end of range */

    this.panel.setRowHeight(    /* height of rows for trains */
      ((Integer)this.rowht.getValue()).intValue());
    try { scale = Double.parseDouble(this.scale.getText()); }
    catch (NumberFormatException e) { scale = 200.0; }
    this.panel.setScale(scale); /* scaling for horizontal axis */
    try { beg   = Double.parseDouble(this.beg.getText()); }
    catch (NumberFormatException e) { beg = Double.NEGATIVE_INFINITY; }
    try { end   = Double.parseDouble(this.end.getText()); }
    catch (NumberFormatException e) { end = Double.POSITIVE_INFINITY; }
    this.setRange(beg, end);    /* beginning and end of range */
    this.panel.repaint();       /* repaint the panel */
  }  /* setLayout() */

  /*------------------------------------------------------------------*/
  /** Get the layout dialog box (create if necessary).
   *  @return the layout dialog box
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog getLayoutDlg ()
  {                             /* --- create layout params. dialog */
    DialogPanel tab;            /* panel for the dialog fields */
    JPanel      bbar;           /* panel for the button bar */
    JButton     button;         /* button for apply and close */

    if (this.layout != null)    /* if the dialog box exists, */
      return this.layout;       /* simply return it */
    this.layout = new JDialog(this);
    tab = new DialogPanel();    /* create the main panel */

    tab.addLabel("Row height [pixels]:");
    this.rowht = tab.addSpinner(2, 1, 100, 1);
    tab.addHelp("Number of pixels for each train (one row per train).");

    tab.addLabel("Scaling [pixels/unit]:");
    tab.add(this.scale = new JTextField("200.0"), DialogPanel.RIGHT);
    tab.addHelp("Scaling factor for the horizontal (time) axis.");

    tab.addLabel("Time range to display:");
    tab.add(this.beg = new JTextField(""), DialogPanel.MIDDLE);
    tab.add(this.end = new JTextField(""), DialogPanel.RIGHT);
    tab.addHelp("Only events/spikes in this (time) range "
               +"are displayed.");

    bbar = new JPanel(new GridLayout(1, 2, 4, 4));
    bbar.setBorder(BorderFactory.createEmptyBorder(0, 4, 4, 4));
    bbar.add(button = new JButton("Apply"));
    button.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.setLayout(); } } );
    bbar.add(button = new JButton("Ok"));
    button.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.layout.setVisible(false);
        TrainView.this.setLayout(); } } );
    bbar.add(button = new JButton("Cancel"));
    button.addActionListener(new ActionListener () {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.layout.setVisible(false); } } );

    this.layout.getContentPane().add(tab,  BorderLayout.CENTER);
    this.layout.getContentPane().add(bbar, BorderLayout.SOUTH);
    this.layout.setTitle("Set Layout...");
    this.layout.setLocationRelativeTo(this);
    this.layout.pack();
    return this.layout;
  }  /* getLayoutDlg() */

  /*------------------------------------------------------------------*/
  /** Set the color of the bars.
   *  @param  c the color of the bars
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setColor (Color c)
  {                             /* --- set the color of the bars */
    this.panel.setColor(c);     /* set the color and */
    this.panel.repaint();       /* repaint the panel */
  }  /* setColor() */

  /*------------------------------------------------------------------*/
  /** Get the color selection dialog box (create if necessary).
   *  @return the color selection dialog box
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private JDialog getColorDlg ()
  {                             /* --- create color chooser dialog */
    if (this.colsel != null)    /* if the dialog box exists, */
      return this.colsel;       /* simply return it */
    this.colsel = new ColorDialog(this, "Set Color...");
    this.colsel.setColor(new Color(0.0F, 0.0F, 0.8F));
    this.colsel.addActionListener(new ActionListener() {
      public void actionPerformed (ActionEvent e) {
        TrainView.this.setColor(TrainView.this.colsel.getColor()); } });
    return this.colsel;         /* return the created color dialog */
  }  /* getColorDlg() */

  /*------------------------------------------------------------------*/
  /** Set the train set to display.
   *  @param  trains the train set to display
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setTrains (TrainSet trains)
  { this.panel.setTrains(trains); }

  /*------------------------------------------------------------------*/
  /** Get the current train set to display.
   *  @return the current train set to display
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public TrainSet getTrains ()
  { return this.panel.getTrains(); }

  /*------------------------------------------------------------------*/
  /** Set the range of event/spike times to display.
   *  @param  beg the beginning of the time range to display
   *  @param  end the end       of the time range to display
   *  @since  2013.11.30 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setRange (double beg, double end)
  {                             /* --- set the range to display */
    this.getLayoutDlg();        /* ensure that layout dialog exists */
    this.beg.setText((beg > Double.NEGATIVE_INFINITY)
                     ? String.valueOf(beg) : "");
    this.end.setText((end < Double.POSITIVE_INFINITY)
                     ? String.valueOf(end) : "");
    this.panel.setRange(beg, end);
  }  /* setRange() */

  /*------------------------------------------------------------------*/
  /** Report an I/O error.
   *  @param  msg the error message to report
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void reportError (String msg)
  {                             /* --- report an i/o error */
    this.status.setText(msg);   /* set message in status bar, */
    System.err.println();       /* print message to stderr, */
    System.err.println(msg);    /* and show alert dialog box */
    JOptionPane.showMessageDialog(this, msg,
      "Error", JOptionPane.ERROR_MESSAGE);
  }  /* reportError() */

  /*------------------------------------------------------------------*/
  /** Load the train set to display.
   *  @param  file the file to load the trains from
   *  @return whether the file was successfully loaded
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean loadTrains (File file)
  {                             /* --- load a train set */
    TrainSet trains;            /* loaded train set */

    if (file == null) {         /* if no file name is given */
      int r = this.getFileChooser().showOpenDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return false;
      file = this.chooser.getSelectedFile();
    }                           /* get the selected file */
    this.getFormatDlg();        /* create format dialog if necessary */
    this.getLayoutDlg();        /* create layout dialog if necessary */
    try {                       /* and create corresp. table reader */
      System.err.print("reading " +file +" ... ");
      long        t      = System.currentTimeMillis();
      TableReader reader = this.format.createReader(file);
      int         mode   = this.format.getTrainMode();
      trains = TrainSet.parse(null, reader, mode);
      reader.close();           /* read trains from the given file */
      this.setTrains(trains);   /* set the loaded trains */
      t = System.currentTimeMillis() -t;
      System.err.print("[" +trains.getSize() +" trains(s)] done");
      System.err.println(" [" +(t/(float)1000.0) +"s]."); }
    catch (IOException e) {     /* catch and report an i/o error */
      this.reportError(e.getMessage()); return false; }
    double d = Math.ceil (trains.getMaxTime())
             - Math.floor(trains.getMinTime());
    if      (d <=   1.0) d = 600.0;
    else if (d <=   2.0) d = 300.0;
    else if (d <=   3.0) d = 200.0;
    else if (d <=   6.0) d = 100.0;
    else if (d <=   8.0) d =  75.0;
    else if (d <=  12.0) d =  50.0;
    else if (d <=  15.0) d =  40.0;
    else if (d <=  20.0) d =  30.0;
    else if (d <=  24.0) d =  25.0;
    else if (d <=  30.0) d =  20.0;
    else if (d <=  40.0) d =  25.0;
    else if (d <=  60.0) d =  10.0;
    else if (d <=  75.0) d =   8.0;
    else if (d <= 100.0) d =   6.0;
    else if (d <= 120.0) d =   5.0;
    else if (d <= 300.0) d =   2.0;
    else                 d =   1.0;
    this.scale.setText(String.valueOf(d));
    this.panel.setScale(d);     /* compute and set scaling factor */
    this.panel.repaint();       /* finally repaint the panel */
    this.status.setText(file.getPath());
    this.curr = file;           /* note the new file name */
    return true;                /* return 'loading successful' */
  }  /* loadTrains() */

  /*------------------------------------------------------------------*/
  /** Save a PNG image of the panel.
   *  @param  file the file to save the image to
   *  @return whether the file was successfully written
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public boolean saveImage (File file)
  {                             /* --- save image to a file */
    if (file == null) {         /* if no file name is given */
      int r = this.getFileChooser().showSaveDialog(this);
      if (r != JFileChooser.APPROVE_OPTION) return false;
      file = this.chooser.getSelectedFile();
    }                           /* let the user choose a file name */
    try {                       /* open an output stream */
      System.err.print("writing " +file +" ... ");
      long             t      = System.currentTimeMillis();
      FileOutputStream stream = new FileOutputStream(file);
      ImageIO.write(this.panel.makeImage(), "png", stream);
      stream.close();           /* save the decision tree image */
      t = System.currentTimeMillis() -t;
      System.err.println("done [" +(t/(float)1000.0) +"s]."); }
    catch (IOException e) {     /* catch and report an i/o error */
      this.reportError(e.getMessage()); return false; }
    return true;                /* return 'ok' */
  }  /* saveImage() */

  /*------------------------------------------------------------------*/
  /** Main function for command line invocation.
   *  @param  args the command line arguments
   *  @since  2013.11.29 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public static void main (String[] args)
  {                             /* --- main function */
    TrainView tv = new TrainView(PROGRAM);
    if (args.length > 0) tv.loadTrains(new File(args[0]));
    tv.setVisible(true);        /* create and show a bar chart */
  }  /* main() */

}  /* class TrainView */
